@@ -15,10 +15,10 @@ use crate::{
     peer::Peer,
     relocation::{RelocateDetails, RelocatePromise, RelocateState},
     routing::command::Command,
-    section::{MemberInfo, Section, SectionAuthorityProvider, SectionChain},
+    section::{ElderCandidates, MemberInfo, Section, SectionChain},
 };
 use bytes::Bytes;
-use sn_messaging::DstLocation;
+use sn_messaging::{DestInfo, DstLocation};
 use std::{cmp::Ordering, iter, net::SocketAddr, slice};
 use xor_name::XorName;
 
@@ -37,6 +37,7 @@ impl Core {
         );
 
         let addr = *member_info.value.peer.addr();
+        let name = *member_info.value.peer.name();
 
         // Attach proof chain that includes the key the approved node knows (if any), the key its
         // `MemberInfo` is signed with and the last key of our section chain.
@@ -52,29 +53,33 @@ impl Core {
             member_info,
         };
 
-        let message = Message::single_src(
-            &self.node,
-            DstLocation::Direct,
-            variant,
-            Some(proof_chain),
-            None,
-        )?;
+        let message =
+            Message::single_src(&self.node, DstLocation::Direct, variant, Some(proof_chain))?;
 
-        Ok(Command::send_message_to_node(&addr, message.to_bytes()))
+        Ok(Command::send_message_to_node(
+            (name, addr),
+            message.to_bytes(),
+            DestInfo {
+                dest: name,
+                dest_section_pk: *self.section.chain().last_key(),
+            },
+        ))
     }
 
     pub(crate) fn send_sync(&mut self, section: Section, network: Network) -> Result<Vec<Command>> {
-        let send = |variant, recipients: Vec<_>| -> Result<_> {
+        let send = |variant, recipients: Vec<(XorName, SocketAddr)>| -> Result<_> {
             trace!("Send {:?} to {:?}", variant, recipients);
 
-            let message =
-                Message::single_src(&self.node, DstLocation::Direct, variant, None, None)?;
-            let recipients: Vec<_> = recipients.iter().map(Peer::addr).copied().collect();
-
+            let message = Message::single_src(&self.node, DstLocation::Direct, variant, None)?;
+            let dest_info = DestInfo {
+                dest: XorName::random(),
+                dest_section_pk: *self.section.chain().last_key(),
+            };
             Ok(Command::send_message_to_nodes(
-                &recipients,
+                recipients.clone(),
                 recipients.len(),
                 message.to_bytes(),
+                dest_info,
             ))
         };
 
@@ -83,8 +88,8 @@ impl Core {
         let (elders, non_elders): (Vec<_>, _) = section
             .active_members()
             .filter(|peer| peer.name() != &self.node.name())
-            .copied()
-            .partition(|peer| section.is_elder(peer.name()));
+            .map(|peer| (*peer.name(), *peer.addr()))
+            .partition(|peer| section.is_elder(&peer.0));
 
         // Send the trimmed state to non-elders. The trimmed state contains only the knowledge of
         // own section.
@@ -106,25 +111,32 @@ impl Core {
         let send = |variant, recipients: Vec<_>| -> Result<_> {
             trace!("Send {:?} to {:?}", variant, recipients);
 
-            let message =
-                Message::single_src(&self.node, DstLocation::Direct, variant, None, None)?;
-            let recipients: Vec<_> = recipients.iter().map(Peer::addr).copied().collect();
+            let message = Message::single_src(&self.node, DstLocation::Direct, variant, None)?;
 
             Ok(Command::send_message_to_nodes(
-                &recipients,
+                recipients.clone(),
                 recipients.len(),
                 message.to_bytes(),
+                DestInfo {
+                    dest: XorName::random(),
+                    dest_section_pk: *self.section_chain().last_key(),
+                },
             ))
         };
 
         let mut commands = vec![];
 
-        let adults: Vec<_> = self.section.live_adults().copied().collect();
+        let adults: Vec<_> = self
+            .section
+            .live_adults()
+            .map(|peer| (*peer.name(), *peer.addr()))
+            .collect();
 
         let variant = Variant::Sync {
             section: self.section.clone(),
             network: Network::new(),
         };
+
         commands.push(send(variant, adults)?);
 
         Ok(commands)
@@ -138,7 +150,7 @@ impl Core {
         // We need to construct a proof that would be trusted by the destination section.
         let known_key = self
             .network
-            .knowledge_by_name(&details.destination)
+            .key_by_name(&details.destination)
             .unwrap_or_else(|| self.section.chain().root_key());
 
         let src = details.pub_id;
@@ -178,34 +190,31 @@ impl Core {
         }
     }
 
-    pub(crate) fn send_dkg_start(
-        &self,
-        section_auth: SectionAuthorityProvider,
-    ) -> Result<Vec<Command>> {
+    pub(crate) fn send_dkg_start(&self, elder_candidates: ElderCandidates) -> Result<Vec<Command>> {
         // Send to all participants.
-        let recipients: Vec<_> = section_auth.peers().collect();
-        self.send_dkg_start_to(section_auth, &recipients)
+        let recipients: Vec<_> = elder_candidates.peers().collect();
+        self.send_dkg_start_to(elder_candidates, &recipients)
     }
 
     pub(crate) fn send_dkg_start_to(
         &self,
-        section_auth: SectionAuthorityProvider,
+        elder_candidates: ElderCandidates,
         recipients: &[Peer],
     ) -> Result<Vec<Command>> {
-        let src_prefix = section_auth.prefix;
+        let src_prefix = elder_candidates.prefix;
         let generation = self.section.chain().main_branch_len() as u64;
-        let dkg_key = DkgKey::new(&section_auth, generation);
+        let dkg_key = DkgKey::new(&elder_candidates, generation);
 
         trace!(
-            "Send DkgStart for {} with {:?} to {:?}",
-            section_auth,
+            "Send DkgStart for {:?} with {:?} to {:?}",
+            elder_candidates,
             dkg_key,
             recipients
         );
 
         let variant = Variant::DkgStart {
             dkg_key,
-            section_auth,
+            elder_candidates,
         };
 
         self.send_message_for_dst_accumulation(
@@ -257,14 +266,6 @@ impl Core {
         recipients: &[Peer],
     ) -> Result<Vec<Command>> {
         let proof_chain = self.create_proof_chain(&dst, additional_proof_chain_key)?;
-        let dst_key = if let Some(name) = dst.name() {
-            *self.section_key_by_name(&name)
-        } else {
-            // NOTE: `dst` is `Direct`. We use this only if the recipient is in our section, so
-            // it's OK to use our latest key as the `dst_key`.
-            *self.section.chain().last_key()
-        };
-
         let key_share = self.section_keys_provider.key_share().map_err(|err| {
             trace!(
                 "Can't create message {:?} for accumulation at dst {:?}: {}",
@@ -274,14 +275,7 @@ impl Core {
             );
             err
         })?;
-        let message = Message::for_dst_accumulation(
-            key_share,
-            src,
-            dst,
-            variant,
-            proof_chain,
-            Some(dst_key),
-        )?;
+        let message = Message::for_dst_accumulation(key_share, src, dst, variant, proof_chain)?;
 
         trace!(
             "Send {:?} for accumulation at dst to {:?}",
@@ -305,15 +299,21 @@ impl Core {
             if recipient.name() == &self.node.name() {
                 handle = true;
             } else {
-                others.push(*recipient.addr());
+                others.push((*recipient.name(), *recipient.addr()));
             }
         }
 
         if !others.is_empty() {
+            let count = others.len();
+            let dest_section_pk = *self.section_key_by_name(&others[0].0);
             commands.push(Command::send_message_to_nodes(
-                &others,
-                others.len(),
+                others,
+                count,
                 message.to_bytes(),
+                DestInfo {
+                    dest: XorName::random(), // will be updated when sending
+                    dest_section_pk,
+                },
             ));
         }
 
@@ -321,6 +321,10 @@ impl Core {
             commands.push(Command::HandleMessage {
                 sender: Some(self.node.addr),
                 message,
+                dest_info: DestInfo {
+                    dest: self.node.name(),
+                    dest_section_pk: *self.section_chain().last_key(),
+                },
             });
         }
 
@@ -345,7 +349,7 @@ impl Core {
         // the actual last key of the resulting proof chain because it's the key that will be used
         // to sign the message.
         let additional_key = additional_key
-            .or_else(|| self.network.knowledge_by_name(&dst.name()?))
+            .or_else(|| self.network.key_by_name(&dst.name()?))
             .filter(|key| self.section.chain().cmp_by_position(key, &last_key) == Ordering::Less);
 
         Ok(self
@@ -356,11 +360,19 @@ impl Core {
 
     pub(crate) fn send_direct_message(
         &self,
-        recipient: &SocketAddr,
+        recipient: (XorName, SocketAddr),
         variant: Variant,
+        dst_pk: bls::PublicKey,
     ) -> Result<Command> {
-        let message = Message::single_src(&self.node, DstLocation::Direct, variant, None, None)?;
-        Ok(Command::send_message_to_node(recipient, message.to_bytes()))
+        let message = Message::single_src(&self.node, DstLocation::Direct, variant, None)?;
+        Ok(Command::send_message_to_node(
+            recipient,
+            message.to_bytes(),
+            DestInfo {
+                dest: recipient.0,
+                dest_section_pk: dst_pk,
+            },
+        ))
     }
 
     // TODO: consider changing this so it sends only to a subset of the elders
@@ -369,10 +381,18 @@ impl Core {
         let targets: Vec<_> = self
             .section
             .authority_provider()
-            .elders
-            .values()
-            .copied()
+            .elders()
+            .iter()
+            .map(|(name, address)| (*name, *address))
             .collect();
-        Command::send_message_to_nodes(&targets, targets.len(), msg)
+
+        let dest_section_pk = *self.section_chain().last_key();
+
+        let dest_info = DestInfo {
+            dest: self.section.prefix().name(),
+            dest_section_pk,
+        };
+
+        Command::send_message_to_nodes(targets.clone(), targets.len(), msg, dest_info)
     }
 }
