@@ -15,7 +15,7 @@ use super::{
     enduser_registry::EndUserRegistry, lazy_messaging, split_barrier::SplitBarrier,
 };
 use crate::{
-    agreement::{DkgVoter, Proposal, ProposalAggregator},
+    agreement::{DkgVoter, Proposal, ProposalAggregator, Proven},
     error::Result,
     event::{Elders, Event, NodeElderChange},
     message_filter::MessageFilter,
@@ -23,17 +23,21 @@ use crate::{
     network::Network,
     node::Node,
     relocation::RelocateState,
-    section::{Section, SectionKeyShare, SectionKeysProvider},
+    section::{Section, SectionAuthorityProvider, SectionKeyShare, SectionKeysProvider},
+    SectionChain,
 };
 use bls_signature_aggregator::SignatureAggregator;
 use itertools::Itertools;
 use resource_proof::ResourceProof;
+use sn_messaging::DestInfo;
+use std::collections::BTreeMap;
 use std::collections::BTreeSet;
+use std::net::SocketAddr;
 use tokio::sync::mpsc;
 use xor_name::{Prefix, XorName};
 
-pub(crate) const RESOURCE_PROOF_DATA_SIZE: usize = 64;
-pub(crate) const RESOURCE_PROOF_DIFFICULTY: u8 = 2;
+pub const RESOURCE_PROOF_DATA_SIZE: usize = 64;
+pub const RESOURCE_PROOF_DIFFICULTY: u8 = 2;
 const KEY_CACHE_SIZE: u8 = 5;
 
 // State + logic of a routing node.
@@ -54,6 +58,13 @@ pub(crate) struct Core {
     resource_proof: ResourceProof,
     end_users: EndUserRegistry,
     connectivity_complaints: ConnectivityComplaints,
+    lagging_messages: LaggingMessages,
+}
+
+#[derive(Default)]
+pub(crate) struct LaggingMessages {
+    // Execute upon sync
+    pub src_ahead: BTreeMap<bls::PublicKey, Vec<Command>>,
 }
 
 impl Core {
@@ -82,6 +93,7 @@ impl Core {
             resource_proof: ResourceProof::new(RESOURCE_PROOF_DATA_SIZE, RESOURCE_PROOF_DIFFICULTY),
             end_users: EndUserRegistry::new(),
             connectivity_complaints: ConnectivityComplaints::new(),
+            lagging_messages: LaggingMessages::default(),
         }
     }
 
@@ -89,24 +101,34 @@ impl Core {
     // Miscellaneous
     ////////////////////////////////////////////////////////////////////////////
 
-    // Update our knowledge of their (sender's) section and their knowledge of our section.
-    pub(crate) fn update_section_knowledge(&mut self, msg: &Message) -> Result<Vec<Command>> {
+    fn check_for_entropy(
+        &mut self,
+        msg: &Message,
+        dest_info: DestInfo,
+        sender: Option<SocketAddr>,
+    ) -> Result<(Vec<Command>, bool)> {
         if !self.is_elder() {
-            return Ok(vec![]);
+            // Adult nodes do need to carry out entropy checking, however the message shall always
+            // be handled.
+            return Ok((vec![], true));
         }
 
-        let actions = lazy_messaging::process(&self.node, &self.section, &self.network, msg)?;
+        let (actions, can_be_executed) = lazy_messaging::process(
+            &self.node,
+            &self.section,
+            &self.network,
+            &mut self.lagging_messages,
+            msg,
+            dest_info,
+            sender,
+        )?;
         let mut commands = vec![];
 
-        if let Some(msg) = actions.send {
+        for msg in actions.send {
             commands.extend(self.relay_message(&msg)?);
         }
 
-        if let Some(proposal) = actions.propose {
-            commands.extend(self.propose(proposal)?);
-        }
-
-        Ok(commands)
+        Ok((commands, can_be_executed))
     }
 
     pub(crate) fn state_snapshot(&self) -> StateSnapshot {
@@ -114,13 +136,31 @@ impl Core {
             is_elder: self.is_elder(),
             last_key: *self.section.chain().last_key(),
             prefix: *self.section.prefix(),
-            elders: self
-                .section()
-                .authority_provider()
-                .elders
-                .keys()
-                .copied()
-                .collect(),
+            elders: self.section().authority_provider().names(),
+        }
+    }
+
+    pub(crate) fn update_section_knowledge(
+        &mut self,
+        section_auth: Proven<SectionAuthorityProvider>,
+        section_chain: SectionChain,
+    ) {
+        let prefix = section_auth.value.prefix;
+        if self
+            .network
+            .update_their_key((prefix, *section_chain.last_key()))
+        {
+            info!("Keys updated for Prefix: {:?}", prefix);
+        } else {
+            info!("Nothing to update, Keys are still the same");
+        }
+        if self
+            .network
+            .update_section(section_auth, None, &section_chain)
+        {
+            info!("Neighbour section knowledge updated: {:?}", prefix);
+        } else {
+            warn!("Neighbour section update failed");
         }
     }
 
@@ -158,8 +198,22 @@ impl Core {
 
             if new.is_elder || old.is_elder {
                 commands.extend(self.send_sync(self.section.clone(), self.network.clone())?);
+                commands.extend(self.handle_lagging_messages_on_sync()?);
             }
 
+            let current: BTreeSet<_> = self.section.authority_provider().names();
+            let added = current.difference(&old.elders).copied().collect();
+            let removed = old.elders.difference(&current).copied().collect();
+            let remaining = old.elders.intersection(&current).copied().collect();
+
+            let elders = Elders {
+                prefix: new.prefix,
+                key: new.last_key,
+                remaining,
+                added,
+                removed,
+            };
+
             let self_status_change = if !old.is_elder && new.is_elder {
                 info!("Promoted to elder");
                 NodeElderChange::Promoted
@@ -172,43 +226,20 @@ impl Core {
                 NodeElderChange::None
             };
 
-            let current: BTreeSet<_> = self
-                .section
-                .authority_provider()
-                .elders
-                .keys()
-                .copied()
-                .collect();
-            let added = current.difference(&old.elders).copied().collect();
-            let removed = old.elders.difference(&current).copied().collect();
-            let remaining = old.elders.intersection(&current).copied().collect();
-
-            let elders = Elders {
-                prefix: new.prefix,
-                key: new.last_key,
-                remaining,
-                added,
-                removed,
-            };
-
             let sibling_elders = if new.prefix != old.prefix {
-                if let Some(sibling_key) = self.section_key(&new.prefix.sibling()) {
-                    self.network.get(&new.prefix.sibling()).map(|info| {
-                        let current: BTreeSet<_> = info.elders.keys().copied().collect();
-                        let added = current.difference(&old.elders).copied().collect();
-                        let removed = old.elders.difference(&current).copied().collect();
-                        let remaining = old.elders.intersection(&current).copied().collect();
-                        Elders {
-                            prefix: new.prefix.sibling(),
-                            key: *sibling_key,
-                            remaining,
-                            added,
-                            removed,
-                        }
-                    })
-                } else {
-                    None
-                }
+                self.network.get(&new.prefix.sibling()).map(|sec_auth| {
+                    let current: BTreeSet<_> = sec_auth.names();
+                    let added = current.difference(&old.elders).copied().collect();
+                    let removed = old.elders.difference(&current).copied().collect();
+                    let remaining = old.elders.intersection(&current).copied().collect();
+                    Elders {
+                        prefix: new.prefix.sibling(),
+                        key: sec_auth.section_key(),
+                        remaining,
+                        added,
+                        removed,
+                    }
+                })
             } else {
                 None
             };
